/*------------------------------------------------------------------------------
 * Copyright (c) 2004-2014 Arm Limited (or its affiliates). All
 * rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *   1.Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   2.Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *   3.Neither the name of Arm nor the names of its contributors may be used
 *     to endorse or promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDERS AND CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *------------------------------------------------------------------------------
 * Name:    Blinky.c
 * Purpose: Dual Core LED Flasher for MCB4300
 *----------------------------------------------------------------------------*/

#include "LPC43xx.h"                    /* LPC43xx Definitions                */
#include "Board_LED.h"

const                                   /* Place the CM0 image in CM4 ROM     */
#include "CM0_Image.c"                  /* Cortex M0 image reference          */

uint32_t LEDOn, LEDOff;

uint32_t M4_EventTimer, M0_EventResp;

/*------------------------------------------------------------------------------
  SysTick IRQ Handler @ 50ms
 *----------------------------------------------------------------------------*/
void SysTick_Handler (void) {
  static uint32_t ticks;
  
  switch (ticks++) {
    case  0: LEDOn  = 1; break;
    case  2: LEDOff = 1; break;
    case  4: ticks  = 0; break;
    
    default:
      if (ticks > 10) {
        ticks = 0;
      }
  }
  if (M4_EventTimer) {
    M4_EventTimer -= 50;
  }
}

/*------------------------------------------------------------------------------
  M0APP Core IRQ Handler
 *----------------------------------------------------------------------------*/
void M0APP_IRQHandler (void) {
  LPC_CREG->M0APPTXEVENT = 0;
  M0_EventResp = 1;
}

/*------------------------------------------------------------------------------
  Load Cortex M0APP Application Image
 *----------------------------------------------------------------------------*/
void Load_CM0_Image (uint32_t DestAddr, const uint8_t *Image, uint32_t Sz) {
  uint32_t i;
  uint8_t *dp = (uint8_t *)DestAddr;

  /* Copy application image */
  for (i = 0; i < Sz; i++) {
    dp[i] = Image[i];
  }
  /* Set shadow pointer to beginning of the CM0 application */
  LPC_CREG->M0APPMEMMAP = DestAddr;
}

/*------------------------------------------------------------------------------
  Main function
 *----------------------------------------------------------------------------*/
int main (void) {
  uint32_t LEDState    = 0;
  uint32_t M4_EventSet = 1;

  SystemCoreClockUpdate ();             /* Update system core clock           */
  SysTick_Config(SystemCoreClock/20);   /* Generate interrupt each 50 ms      */
  LED_Initialize();                     /* LED Initialization                 */
  NVIC_EnableIRQ (M0APP_IRQn);          /* Enable IRQ from the M0APP Core     */

  /* Stop CM0 core */
  LPC_RGU->RESET_CTRL1 = (1 << 24);

  /* Download CM0 application */
  Load_CM0_Image (0x10080000, LR0, sizeof (LR0)); 
  
  /* Start CM0 core */
  LPC_RGU->RESET_CTRL1 = 0;

  while (1) {
    if (LEDOn) {
      LEDOn = 0;
      LED_On (0);                       /* Turn on LED 0                      */
    }

    if (LEDOff) {
      LEDOff = 0;
      LED_Off (0);                      /* Turn off LED 0                     */
    }
    
    if (M4_EventSet) {
      M4_EventSet = 0;
      __sev();                          /* Send event signal to the CM0 core  */
    }

    if (M4_EventTimer == 0) {           /* Wait for timer timeout (250ms)     */
      if (M0_EventResp) {
        M0_EventResp = 0;               /* M0 response received, toggle LED 3 */

        (LEDState ^= 1) ? LED_On (3) : LED_Off (3);

        M4_EventSet =   1;
        M4_EventTimer  = 250;
      }
    }
  }
}
